/*********************************************************************

Author:     Tomas Franzon
Date:       2004-06-15
Program:    UserPort.SYS
Compile:    Use DDK BUILD facility

Purpose:    Give direct port I/O access to user mode programs.

This driver is influenced by an article written by Dale Roberts 8/30/95,
published in May 96 Dr Dobbs Journal, see www.ddj.com.
The driver gives user mode program access to selected ports by changing
the x86-processors IOPM (I/O Permission Map). The driver has one separate
map for all procsesses (AllProcessesIOPM) and one separate map 
(ThroughCreateFileIOPM) for processes that opens the file "\\.\UserPort"

The driver tries to read the registry keys:
HKEY_LOCAL_MACHINE\Software\UserPort\AllProcessesIOPM
HKEY_LOCAL_MACHINE\Software\UserPort\ThroughCreateFileIOPM
It will use the default values below if these doesn't exist.

History:
Version 1.0: 
*First release
Version 1.1:
* Added Test LPT1 button
* Added Test Spkr button
* Start button is greyed when the driver is started
* Stop button is greyed when the driver is stopped
* UserPort driver now works on Multiprocessor machines and Hyperthreading machines
* UserPort now works on all Windows XP machines
* \\.\UserPort access is now stopped when the \\.\UserPort file is closed
* Sleep(100) is no longer needed after the CreateFile("\\.\UserPort"...) call
* Address range is extended to 0x7a80 for all processes and to 0xffff for \\.\UserPort processes
* Added Java and VisualBasic example files

Todo:
Extend the address range to 0xffff for all processes. To do this I need to allocate memory at
address 0x80045000 but I don't know how to do this. Another method would be to change
the default IOPM ponter from 0x20AC to 0x88 for all processes. If you have any Idea about
how to do this please send me an e-mail to tomas_franzon@hotmail.com


*********************************************************************/
#include <ntddk.h>

#define NT4DDK

/*
 *  Make sure our structure is packed properly, on byte boundary, not
 * on the default doubleword boundary.
*/
#pragma pack(push,1)

/*
 *  Structures for manipulating the GDT register and a GDT segment
 * descriptor entry.  Documented in Intel processor handbooks.
 */
typedef struct {
  unsigned limit : 16;
  unsigned baselo : 16;
  unsigned basemid : 8;
  unsigned type : 4;
  unsigned system : 1;
  unsigned dpl : 2;
  unsigned present : 1;
  unsigned limithi : 4;
  unsigned available : 1;
  unsigned zero : 1;
  unsigned size : 1;
  unsigned granularity : 1;
  unsigned basehi : 8;
} GDTENT;

typedef struct {
  unsigned short limit;
  GDTENT *base;
} GDTREG;

#pragma pack(pop)
/*
 *  The name of our device driver.
 */
#define DEVICE_NAME_STRING	L"UserPort"

/*
  OriginalMapCopy is used to restore the IOPM when the driver exists.
  CurrentMap points to NULL or to the place where the processors IOPM is located.
  Accessmap is the IOPM that is used by the driver.
  Every port address has one cooresponding bit in the IOPM. The driver supports
  addresses from 0x000 to 0x3FF and the IOPM size is then 0x3ff / 8 = 0x80.
 */

UCHAR OriginalAllProcIOPMCopy[0x2004];
UCHAR OriginalThroughCreateFileIOPMCopy[0x2004];
const UCHAR DefaultMap[0x80]=
/*0x000*/ {0xFF,0xFF,
/*0x010*/  0xFF,0xFF,
/*0x020*/  0xFF,0xFF,
/*0x030*/  0xFF,0xFF,
/*0x040*/  0xFF,0xFF,
/*0x050*/  0xFF,0xFF,
/*0x060*/  0xFF,0xFF,
/*0x070*/  0xFF,0xFF,
/*0x080*/  0xFF,0xFF,
/*0x090*/  0xFF,0xFF,
/*0x0A0*/  0xFF,0xFF,
/*0x0B0*/  0xFF,0xFF,
/*0x0C0*/  0xFF,0xFF,
/*0x0D0*/  0xFF,0xFF,
/*0x0E0*/  0xFF,0xFF,
/*0x0F0*/  0xFF,0xFF,
/*0x100*/  0xFF,0xFF,
/*0x110*/  0xFF,0xFF,
/*0x120*/  0xFF,0xFF,
/*0x130*/  0xFF,0xFF,
/*0x140*/  0xFF,0xFF,
/*0x150*/  0xFF,0xFF,
/*0x160*/  0xFF,0xFF,
/*0x170*/  0xFF,0xFF,
/*0x180*/  0xFF,0xFF,
/*0x190*/  0xFF,0xFF,
/*0x1A0*/  0xFF,0xFF,
/*0x1B0*/  0xFF,0xFF,
/*0x1C0*/  0xFF,0xFF,
/*0x1D0*/  0xFF,0xFF,
/*0x1E0*/  0xFF,0xFF,
/*0x1F0*/  0xFF,0xFF,
/*0x200*/  0x00,0x00,
/*0x210*/  0x00,0x00,
/*0x220*/  0x00,0x00,
/*0x230*/  0x00,0x00,
/*0x240*/  0x00,0x00,
/*0x250*/  0x00,0x00,
/*0x260*/  0x00,0x00,
/*0x270*/  0x00,0x00,
/*0x280*/  0x00,0x00,
/*0x290*/  0x00,0x00,
/*0x2A0*/  0x00,0x00,
/*0x2B0*/  0x00,0x00,
/*0x2C0*/  0x00,0x00,
/*0x2D0*/  0x00,0x00,
/*0x2E0*/  0x00,0x00,
/*0x2F0*/  0x00,0x00,
/*0x300*/  0x00,0x00,
/*0x310*/  0x00,0x00,
/*0x320*/  0x00,0x00,
/*0x330*/  0x00,0x00,
/*0x340*/  0x00,0x00,
/*0x350*/  0x00,0x00,
/*0x360*/  0x00,0x00,
/*0x370*/  0x00,0x00,
/*0x380*/  0xFF,0xFF,
/*0x390*/  0xFF,0xFF,
/*0x3A0*/  0xFF,0xFF,
/*0x3B0*/  0xFF,0x0F,
/*0x3C0*/  0xFF,0xFF,
/*0x3D0*/  0xFF,0xFF,
/*0x3E0*/  0xFF,0x00,
/*0x3F0*/  0x00,0x00};

unsigned OrgGDTSize;     // The original sise of the TSS
unsigned nMaxAllProcessesIOPM;
unsigned nMaxThroughCreateFileIOPM;
unsigned IOPMPtr;
UCHAR *TSSAllProcessesIOPM;


/*
 Ke386IoSetAccessProcess() adjusts the IOPM offset pointer to the IOPM at 0x88
 Ke386IoSetAccessProcess() is located in NTOSKRNL.EXE but is not included in any
 header file or documented anywhere...
*/

void Ke386IoSetAccessProcess(PEPROCESS, int);
void Ke386SetIoAccessMap(int, UCHAR *);
void Ke386QueryIoAccessMap(int, UCHAR *);

NTSYSAPI
NTSTATUS
NTAPI
ZwYieldExecution (
    VOID
);


/*********************************************************************
  Service handler for a CreateFile() user mode call.

  This routine is entered in the driver object function call table by
the DriverEntry() routine.  When the user mode application calls
CreateFile(), this routine gets called while still in the context of
the user mode application, but with the CPL (the processor's Current
Privelege Level) set to 0.  This allows us to do kernel mode
operations.  UserPort() is called to give the calling process I/O
access.  All the user mode application needs do to obtain I/O access
is open this device with CreateFile().  No other operations are
required.
*********************************************************************/
NTSTATUS CreateFileDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
  if (nMaxThroughCreateFileIOPM != 0) {
    // Give the current process IO access
    Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
    ZwYieldExecution(); // Force a process switch
  }

  Irp->IoStatus.Information = 0;
  Irp->IoStatus.Status = STATUS_SUCCESS;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

NTSTATUS CloseFileDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
  if (nMaxThroughCreateFileIOPM != 0) {
    // Deny the current process IO access
    Ke386IoSetAccessProcess(PsGetCurrentProcess(), 0);
    ZwYieldExecution(); // Force a process switch
  }

  Irp->IoStatus.Information = 0;
  Irp->IoStatus.Status = STATUS_SUCCESS;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// remove the link \\.\UserPort and restore the IOPM 
void UserPortUnload(IN  PDRIVER_OBJECT  DriverObject)
{
  WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
  UNICODE_STRING uniDOSString;
  GDTREG gdtreg;
  GDTENT *g;
  short TaskSeg;
  unsigned i;
  unsigned int NoOfProc;
  KAFFINITY AffinityMask;

  if (nMaxAllProcessesIOPM != 0) {
#ifdef NT4DDK
    NoOfProc = *KeNumberProcessors;
#else
    NoOfProc = KeNumberProcessors;
#endif
    if (NoOfProc < 1)
      NoOfProc = 1;
    if (NoOfProc > 32)
      NoOfProc = 32;
    for (i = 0; i < NoOfProc; i++) {
      // run the "ltr" instruction in every processor on multi processor or hyper threading systems
      AffinityMask = 1 << i;
      ZwSetInformationThread(NtCurrentThread(), ThreadAffinityMask,&AffinityMask, sizeof(KAFFINITY));
      _asm cli;
      _asm str TaskSeg;
      _asm sgdt gdtreg;
      g = gdtreg.base + (TaskSeg >> 3);
      g->limit = OrgGDTSize;
      g->type = 9;
      _asm ltr TaskSeg;
      _asm sti;           
    }
  }

  if (nMaxThroughCreateFileIOPM != 0) {
    // Restore to the original map
    Ke386SetIoAccessMap(1, OriginalThroughCreateFileIOPMCopy);
  }
  if (nMaxAllProcessesIOPM != 0) {
    for (i=0;i<nMaxAllProcessesIOPM + 4;i++) {
      TSSAllProcessesIOPM[i] = OriginalAllProcIOPMCopy[i];
    }
  }

  RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);
  IoDeleteSymbolicLink (&uniDOSString);
  IoDeleteDevice(DriverObject->DeviceObject);
}

// This routine is the entrypoint of the driver.
// This routine reads the AllProcessesIOPM and ThroughCreateFileIOPM from registry and start the driver
NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
  PDEVICE_OBJECT deviceObject;
  NTSTATUS status;
  WCHAR NameBuffer[] = L"\\Device\\" DEVICE_NAME_STRING;
  WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
  UNICODE_STRING uniNameString, uniDOSString;
  GDTREG gdtreg;
  GDTENT *g;
  short TaskSeg;
  UCHAR *AllProcessesIOPM;
  UCHAR *ThroughCreateFileIOPM;
  unsigned i;
  UCHAR *TSSbase; 
  PMDL pmdl;
  unsigned int NoOfProc;
  KAFFINITY AffinityMask;
  int tst=0;
  int tst2=0;


  PHYSICAL_ADDRESS  HighestAcceptableAddress;
  ULONG nInformation1,nInformation2;
  PKEY_VALUE_PARTIAL_INFORMATION Information1;
  PKEY_VALUE_PARTIAL_INFORMATION Information2;
  ULONG ResultLength;
  HANDLE KeyHandle;
  OBJECT_ATTRIBUTES ObjectAttributes;
  UNICODE_STRING ThroughCreateFileIOPMString,AllProcessesIOPMString,RegPathString;

  RtlInitUnicodeString(&AllProcessesIOPMString, L"AllProcessesIOPM");
  RtlInitUnicodeString(&ThroughCreateFileIOPMString, L"ThroughCreateFileIOPM");
  RtlInitUnicodeString(&RegPathString,L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\UserPort");

  InitializeObjectAttributes(&ObjectAttributes,&RegPathString,OBJ_CASE_INSENSITIVE,NULL,NULL);

  nMaxThroughCreateFileIOPM = 0x80;
  nMaxAllProcessesIOPM = 0x80;
  HighestAcceptableAddress.LowPart = 0xffffffff;
  HighestAcceptableAddress.HighPart = 0xffffffff;
  nInformation1 = sizeof(OriginalThroughCreateFileIOPMCopy)+sizeof(KEY_VALUE_PARTIAL_INFORMATION);
  nInformation2 = sizeof(OriginalAllProcIOPMCopy)+sizeof(KEY_VALUE_PARTIAL_INFORMATION);

  Information1 = (PKEY_VALUE_PARTIAL_INFORMATION)MmAllocateContiguousMemory(nInformation1, HighestAcceptableAddress);
  Information2 = (PKEY_VALUE_PARTIAL_INFORMATION)MmAllocateContiguousMemory(nInformation2, HighestAcceptableAddress);

  RtlFillMemory(Information1, nInformation1, 0xff);
  RtlFillMemory(Information2, nInformation2, 0xff);
  ThroughCreateFileIOPM = Information1->Data;
  AllProcessesIOPM = Information2->Data;

  if (STATUS_SUCCESS == (tst2=ZwOpenKey(&KeyHandle, KEY_QUERY_VALUE,&ObjectAttributes)))	{
    if (STATUS_SUCCESS != ZwQueryValueKey(KeyHandle,&ThroughCreateFileIOPMString,KeyValuePartialInformation,Information1,nInformation1,&ResultLength)) {
      RtlCopyMemory(ThroughCreateFileIOPM, DefaultMap, sizeof(DefaultMap));
      tst|=1;
    }

    if (STATUS_SUCCESS != ZwQueryValueKey(KeyHandle,&AllProcessesIOPMString,KeyValuePartialInformation,Information2,nInformation2,&ResultLength)) {
      RtlCopyMemory(AllProcessesIOPM, DefaultMap, sizeof(DefaultMap));
      tst|=2;
    }
    ZwClose(KeyHandle);
  }
  else {
    RtlCopyMemory(ThroughCreateFileIOPM, DefaultMap, sizeof(DefaultMap));
    RtlCopyMemory(AllProcessesIOPM, DefaultMap, sizeof(DefaultMap));
    tst|=4;
  }

  nMaxThroughCreateFileIOPM = sizeof(OriginalThroughCreateFileIOPMCopy) - 4;
  for (i=sizeof(OriginalThroughCreateFileIOPMCopy) - 1 ; (i >= 0) && ThroughCreateFileIOPM[i] == 0xff ; i--) {
    nMaxThroughCreateFileIOPM = i;
  }

  nMaxAllProcessesIOPM = sizeof(OriginalAllProcIOPMCopy) - 4;
  for (i=sizeof(OriginalAllProcIOPMCopy) - 1 ; (i >= 0) && AllProcessesIOPM[i] == 0xff ; i--) {
    nMaxAllProcessesIOPM = i;
  }


  _asm cli;                           // don't get interrupted!
  _asm str TaskSeg;                   // get the TSS selector      
  _asm sgdt gdtreg;                   // get the GDT address       
  g = gdtreg.base + (TaskSeg >> 3);   // get the TSS descript
  OrgGDTSize = g->limit;
  _asm sti;                           // let interrupts continue

  TSSbase = (UCHAR *) (g->baselo | (g->basemid << 16) | (g->basehi << 24));                
  IOPMPtr = *((USHORT *)(TSSbase + 0x66));
  if (IOPMPtr != 0x20AD) { 
    // IOPMPtr is 0x20AC on Windows XP/2000 and 0x20AD on Windows NT4/3.51
    // but some drivers (video??) on Windows XP sets this pointer incorrectly so 
    // the safest way is to use the hard coded value 0x20AC when not running Windows NT
    IOPMPtr = 0x20AC; 
  }


  // Copy the AccessMap to TSSbase + 0x88 and save the original map
  if (nMaxThroughCreateFileIOPM != 0) {
    Ke386QueryIoAccessMap(1, OriginalThroughCreateFileIOPMCopy);

    for (i=0;i<sizeof(OriginalThroughCreateFileIOPMCopy);i++) {
      // There might be an other driver using the IOPM so lets merge them together
      ThroughCreateFileIOPM[i] = OriginalThroughCreateFileIOPMCopy[i] & ThroughCreateFileIOPM[i];    
    }

    Ke386SetIoAccessMap(1, ThroughCreateFileIOPM);
  }

  TSSAllProcessesIOPM = TSSbase + IOPMPtr;

  if (!MmIsAddressValid(TSSbase + 0x3000-1)) {
    // This OS seem to be different from what is expected. UserPort will not be supported for all processes
    nMaxAllProcessesIOPM = 0;
  }

  // Copy the AccessMap to TSSbase + 0x20ac and save the original map
  if (nMaxAllProcessesIOPM > 0x2ffc - IOPMPtr) {
    nMaxAllProcessesIOPM = 0x2ffc - IOPMPtr;
  }
  RtlCopyMemory(OriginalAllProcIOPMCopy, TSSAllProcessesIOPM, nMaxAllProcessesIOPM);
  RtlCopyMemory(TSSAllProcessesIOPM, AllProcessesIOPM, nMaxAllProcessesIOPM);

  MmFreeContiguousMemory(Information1);
  MmFreeContiguousMemory(Information2);

  if (nMaxAllProcessesIOPM != 0) {
#ifdef NT4DDK
    NoOfProc = *KeNumberProcessors;
#else
    NoOfProc = KeNumberProcessors;
#endif
    if (NoOfProc < 1)
      NoOfProc = 1;
    if (NoOfProc > 32)
      NoOfProc = 32;
    for (i = 0; i < NoOfProc; i++) {
      // run the "ltr" instruction in every processor on multi processor or hyper threading systems
      AffinityMask = 1 << i;
      ZwSetInformationThread(NtCurrentThread(), ThreadAffinityMask,&AffinityMask, sizeof(KAFFINITY));
      _asm cli;                           // don't get interrupted!
      _asm str TaskSeg;                   // get the TSS selector      
      _asm sgdt gdtreg;                   // get the GDT address       
      g = gdtreg.base + (TaskSeg >> 3);   // get the TSS descript
      g->limit = IOPMPtr + nMaxAllProcessesIOPM + 4;  // modify TSS segment limit. 0x3000 is the maximum size (three pages) we can use
      g->type = 9;                        // mark TSS as "not busy"
      _asm ltr TaskSeg;                   // reload task register (TR)
      _asm sti;                           // let interrupts continue
    }
  }

  //
  //  Set up device driver name and device object.
  //  Make the driver accessable though the file \\.\UserPort
	
  RtlInitUnicodeString(&uniNameString, NameBuffer);
  RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);

  status = IoCreateDevice(DriverObject, 0,
                          &uniNameString,
                          FILE_DEVICE_UNKNOWN,
                          0, FALSE, &deviceObject);
  if(!NT_SUCCESS(status))
    return status;

  status = IoCreateSymbolicLink (&uniDOSString, &uniNameString);

  if (!NT_SUCCESS(status))
    return status;

  //
  //  Initialize the Driver Object with driver's entry points.
  // All we require are the Create and Unload operations.
  //
  DriverObject->MajorFunction[IRP_MJ_CREATE] = CreateFileDispatch;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = CloseFileDispatch;

  DriverObject->DriverUnload = UserPortUnload;

  return STATUS_SUCCESS;
}

